package com.fasterxml.jackson.dataformat.avro.deser;

import java.io.IOException;

import org.apache.avro.io.BinaryDecoder;

import com.fasterxml.jackson.core.JsonToken;
import com.fasterxml.jackson.dataformat.avro.AvroReadContext;

public final class MapReader extends AvroStructureReader
{
    protected final static int STATE_START = 0;
    protected final static int STATE_NAME = 1;
    protected final static int STATE_VALUE = 2;
    protected final static int STATE_END = 3;
    protected final static int STATE_DONE = 4;

    private final AvroScalarReader _scalarReader;
    private final AvroStructureReader _structureReader;
    protected final BinaryDecoder _decoder;
    protected final AvroParserImpl _parser;

    private String _currentName;
    
    protected int _state;
    protected long _index;
    protected long _count;
    
    public MapReader(AvroScalarReader reader) {
        this(null, reader, null, null, null);
    }

    public MapReader(AvroStructureReader reader) {
        this(null, null, reader, null, null);
    }
    
    private MapReader(AvroReadContext parent,
            AvroScalarReader scalarReader,
            AvroStructureReader structReader,
            BinaryDecoder decoder, AvroParserImpl parser) {
        super(parent, TYPE_OBJECT);
        _scalarReader = scalarReader;
        _structureReader = structReader;
        _decoder = decoder;
        _parser = parser;
    }
    
    @Override
    public MapReader newReader(AvroReadContext parent,
            AvroParserImpl parser, BinaryDecoder decoder) {
        return new MapReader(parent, _scalarReader, _structureReader, decoder, parser);
    }

    @Override
    public String getCurrentName() { return _currentName; }
    
    @Override
    public JsonToken nextToken() throws IOException
    {
        switch (_state) {
        case STATE_START:
            _parser.setAvroContext(this);
            _count = _decoder.readArrayStart();
            _state = (_count > 0) ? STATE_NAME : STATE_END;
            return JsonToken.START_OBJECT;
        case STATE_NAME:
            if (_index < _count) {
                _currentName = _decoder.readString();
                return JsonToken.FIELD_NAME;
            }
            // need more data...
            _count = _decoder.arrayNext();
            // more stuff?
            if (_count > 0L) {
                _index = 0;
                _currentName = _decoder.readString();
                return JsonToken.FIELD_NAME;
            }
            // otherwise fall through:
        case STATE_END:
            _state = STATE_DONE;
            _parser.setAvroContext(getParent());
            return JsonToken.END_OBJECT;
        case STATE_VALUE:
            break;
        case STATE_DONE:
        default:
            throwIllegalState(_state);
        }
        _state = STATE_NAME;
        ++_index;
        if (_scalarReader != null) {
            return _scalarReader.readValue(_parser, _decoder);
        }
        AvroStructureReader r = _structureReader.newReader(this, _parser, _decoder);
        _parser.setAvroContext(r);
        return r.nextToken();
    }        

    @Override
    public void appendDesc(StringBuilder sb)
    {
        sb.append('{');
        if (_currentName != null) {
            sb.append('"');
            sb.append(_currentName);
            sb.append('"');
        } else {
            sb.append('?');
        }
        sb.append('}');
    }
}